import argparse
import datetime
import time
from pathlib import Path

try:
    from dozent.dozent import Dozent
except ModuleNotFoundError:
    from dozent import Dozent

CURRENT_FILE_PATH = Path(__file__)
DEFAULT_DATA_DIRECTORY = CURRENT_FILE_PATH.parent.parent / "data"

parser = argparse.ArgumentParser(
    description="A powerful downloader to get tweets from twitter for our compute. "
    "The first step of many"
)
parser.add_argument(
    "-s",
    "--start-date",
    help="The date from where we download. The format must be: YYYY-MM-DD",
    type=lambda s: datetime.datetime.strptime(s, "%Y-%m-%d").date(),
)
parser.add_argument(
    "-e",
    "--end-date",
    help="The last day that we download. The format must be: YYYY-MM-DD",
    type=lambda s: datetime.datetime.strptime(s, "%Y-%m-%d").date(),
)
parser.add_argument("-t", "--timeit", help="Show total program runtime", default=True)
parser.add_argument(
    "-o",
    "--output-directory",
    help="Output Directory where the file will be stored. "
    "Defaults to the data/ directory",
    default=DEFAULT_DATA_DIRECTORY,
)
parser.add_argument(
    "-q",
    "--quiet",
    help="Turn off output (except for errors and warnings)",
    action="store_true",
)
parser.add_argument(
    "--dry-run",
    help="Downloads 4 small files for testing purposes (<3 MB total)",
    action="store_true",
)
args = parser.parse_args()
command_line_arguments = vars(args)

if __name__ == "__main__":

    _start_time = time.time()
    verbose = not command_line_arguments["quiet"]
    _dozent_object = Dozent()

    if command_line_arguments["start_date"] and command_line_arguments["end_date"]:
        _dozent_object.download_timeframe(
            start_date=command_line_arguments["start_date"],
            end_date=command_line_arguments["end_date"],
            verbose=verbose,
        )

        if command_line_arguments["timeit"]:
            print(
                f"\nDownload Time: {datetime.timedelta(seconds=(time.time() - _start_time))}"
            )

    elif command_line_arguments["dry_run"]:
        _dozent_object.download_test(verbose=verbose)

        if command_line_arguments["timeit"]:
            print(
                f"\n\nDownload Time: {datetime.timedelta(seconds=(time.time() - _start_time))}"
            )

    else:
        parser.print_help()



import datetime
import json
import os
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import List, Dict

try:
    from dozent.downloader_tools import DownloaderTools
except ModuleNotFoundError:
    from downloader_tools import DownloaderTools

CURRENT_FILE_PATH = Path(__file__)
DEFAULT_DATA_DIRECTORY = CURRENT_FILE_PATH.parent.parent / "data"
TWITTER_ARCHIVE_STREAM_LINKS_PATH = (
    CURRENT_FILE_PATH.parent / "twitter-archive-stream-links.json"
)

FIRST_DAY_OF_SUPPORT = datetime.date(2017, 6, 1)
LAST_DAY_OF_SUPPORT = datetime.date(2020, 6, 30)


class _DownloadWorker(Thread):  # skip_tests
    def __init__(
        self,
        queue: Queue,
        download_dir: Path,
        task_id: int,
        number_of_dates: int,
        verbose: bool,
    ):
        Thread.__init__(self)
        self.queue = queue
        self.download_dir = download_dir
        self.task_id = task_id
        self.number_of_dates = number_of_dates
        self.verbose = verbose

    def run(self):
        while True:
            # Get the work from the queue and expand the tuple
            link = self.queue.get()
            try:
                DownloaderTools.download_with_pysmartdl(
                    link=link,
                    download_dir=str(self.download_dir),
                    task_id=self.task_id,
                    number_of_dates=self.number_of_dates,
                    verbose=self.verbose,
                )
            finally:
                self.queue.task_done()


class Dozent:
    # TODO: Move start_date and end_date as optional arguments
    __instance__ = None

    def __init__(self):
        if Dozent.__instance__ is None:
            Dozent.__instance__ = self
            with open(TWITTER_ARCHIVE_STREAM_LINKS_PATH) as file:
                self.date_links: List[Dict[str, str]] = json.loads(file.read())

        else:
            raise RuntimeError('Multiple classes detected, this class might not be thread safe')

    @staticmethod
    def _make_date_from_date_link(date_link: Dict[str, str]) -> datetime.date:
        """
        Function to make a date given the date link. For months that don't have individual dates (date = 'NaN'), we
        assume that it's the first.
        :param date_link: date dictionary from `data/test_sample_files.json`. This is used to convert to the date
        :return: date object for the associated month
        """
        day = int(date_link["day"]) if date_link["day"] != "NaN" else 1
        month = int(date_link["month"])
        year = int(date_link["year"])

        return datetime.date(year=year, month=month, day=day)

    def get_links_for_days(
        self, start_date: datetime.date, end_date: datetime.date
    ) -> List:
        """
        Function to get the links for the given start and end days
        :return: date dictionaries that are within self.start_date and self.end_dates
        """
        links = []

        if not FIRST_DAY_OF_SUPPORT >= start_date:
            RuntimeError(
                f'We currently only support the range {FIRST_DAY_OF_SUPPORT.strftime("%d, %b %Y")}, '
                f'{LAST_DAY_OF_SUPPORT.strftime("%d, %b %Y")}'
                f"\nWe're planning on adding support for it soon."
                f"Need that data sooner? Add an issue to out GitHub repo and we'll "
                f"walk you through the process"
                f"Link to our GitHub "
                f"Issues: https://github.com/Twitter-Public-Analysis/Twitter-Public-Analysis/issues"
            )

        elif not end_date <= end_date:
            RuntimeError(
                f'We currently only support the range {FIRST_DAY_OF_SUPPORT.strftime("%d, %b %Y")}, '
                f'{LAST_DAY_OF_SUPPORT.strftime("%d, %b %Y")}'
                f"\nWe're planning on adding support for it soon."
                f"Need that data sooner? Add an issue to out GitHub repo and we'll "
                f"walk you through the process"
                f"Link to our GitHub "
                f"Issues: https://github.com/Twitter-Public-Analysis/Twitter-Public-Analysis/issues"
            )

        for date_link in self.date_links:
            date = Dozent._make_date_from_date_link(date_link)
            if start_date <= date <= end_date:
                links.append(date_link)

        return links

    def download_timeframe(
        self,
        start_date: datetime.date,
        end_date: datetime.date,
        verbose: bool = True,
        download_dir: Path = DEFAULT_DATA_DIRECTORY,
    ):  # skip_tests
        """
        Download all tweet archives from self.start_date to self.end_date
        :return: None
        """

        # Create a queue to communicate with the worker threads
        queue = Queue()

        os.makedirs(download_dir, exist_ok=True)

        number_of_dates = (end_date - start_date).days + 1

        task_id = 0

        for x in range(number_of_dates):
            worker = _DownloadWorker(
                queue=queue,
                download_dir=download_dir,
                task_id=task_id,
                number_of_dates=number_of_dates,
                verbose=verbose,
            )
            task_id += 1
            # Setting daemon to True will let the main thread exit even though the workers are blocking
            worker.daemon = True
            worker.start()

        for sample_date in self.get_links_for_days(
            start_date=start_date, end_date=end_date
        ):
            print(
                f"Queueing tweets download for {sample_date['day']}-{sample_date['month']}-{sample_date['year']}"
            )
            queue.put(sample_date["link"])

        print("")
        queue.join()

    def download_test(
        self, verbose: bool = True, download_dir: Path = DEFAULT_DATA_DIRECTORY
    ):  # skip_tests
        """
        Downloads four small test files from S3 for testing purposes
        """

        # Stores download links for sample data
        test_download_links = [
            "https://dozent-tests.s3.amazonaws.com/test_500K.txt",
            "https://dozent-tests.s3.amazonaws.com/test_550K.txt",
            "https://dozent-tests.s3.amazonaws.com/test_600K.txt",
            "https://dozent-tests.s3.amazonaws.com/test_650K.txt",
        ]

        # Create a queue to communicate with the worker threads
        queue = Queue()

        os.makedirs(download_dir, exist_ok=True)

        task_id = 0

        number_of_links = len(test_download_links)

        for x in range(number_of_links):
            worker = _DownloadWorker(
                queue=queue,
                download_dir=download_dir,
                task_id=task_id,
                number_of_dates=number_of_links,
                verbose=verbose,
            )

            task_id += 1
            # Setting daemon to True will let the main thread exit even though the workers are blocking
            worker.daemon = True
            worker.start()

        for link in test_download_links:
            print(f"Queueing Link {link}")
            queue.put(link)

        print("")
        queue.join()



import random
import sys
import time
from threading import Lock

import humanize
from pySmartDL import SmartDL

_TWO_DECIMALS = "%.2f"

# Used for tracking and displaying download progress bars
global global_progress_tracker
global_progress_tracker = [[0, 0, 0, 0, 0]]

global global_download_size
global_download_size = 0

global global_final_download_size
global_final_download_size = 0

global global_download_speed
global_download_speed = 0

global global_progress_percentage
global_progress_percentage = 0

global global_eta
global_eta = 0


class DownloaderTools:
    __instance__ = None

    def __init__(self):
        if DownloaderTools.__instance__ is None:
            DownloaderTools.__instance__ = self
        else:
            raise RuntimeError(
                f"Singleton {self.__class__.__name__} class is created more than once!"
            )

    @staticmethod
    def _get_individual_download_stats(
        downloader_obj: SmartDL,
    ) -> [int, int, str, float, str]:

        download_size = downloader_obj.get_dl_size()
        file_size = downloader_obj.get_final_filesize()
        speed = downloader_obj.get_speed()
        progress_percentage = downloader_obj.get_progress()
        eta = downloader_obj.get_eta()
        return [download_size, file_size, speed, progress_percentage, eta]

    @staticmethod
    def _update_global_download_size() -> None:
        """
        Updates download progress for all downloads
        """
        global global_download_size
        updated_size = sum([index[0] for index in global_progress_tracker])

        # Prevents lagging threads from updating with lower values
        if updated_size > global_download_size:
            global_download_size = updated_size

    @staticmethod
    def _get_final_download_size() -> int:
        """
        Retrieves the total download size for all downloads
        :return: total download size in bytes
        """
        global_final_download_size = 0
        for download in global_progress_tracker:
            global_final_download_size += download[1]
        return global_final_download_size

    @staticmethod
    def _update_global_download_speed() -> None:
        """
        Updates global download speed for all downloads
        """
        global global_download_speed

        download_speeds = [int(index[2]) for index in global_progress_tracker]
        global_download_speed = int(
            sum(download_speeds) / len(download_speeds)
        )

    @staticmethod
    def _update_global_progress_percentage() -> None:
        """
        Updates global progress percentage for all downloads
        """
        global global_progress_percentage

        updated_percentage = [
            float(index[3]) for index in global_progress_tracker
        ]
        updated_percentage = round(
            float(sum(updated_percentage) / len(updated_percentage) * 100), 2
        )

        # Prevents lagging threads from updating with lower values
        if updated_percentage > global_progress_percentage:
            global_progress_percentage = updated_percentage

    @staticmethod
    def _create_progress_bar(size: int) -> str:
        """
        Creates a progress bar
        :param size: total length of progress bar
        :return: Progress bar as a string
        """
        global global_progress_percentage
        number_of_bars = int(global_progress_percentage * 0.01 * size)
        number_of_dashes = size - number_of_bars

        return f"[{'#' * number_of_bars}{'-' * number_of_dashes}]"

    @staticmethod
    def _update_global_eta() -> None:
        """
        Updates the global eta for total download
        """
        global global_eta

        updated_eta = [float(index[4]) for index in global_progress_tracker]
        updated_eta = int(sum(updated_eta) / len(updated_eta))

        global_eta = updated_eta

    @classmethod
    def download_with_pysmartdl(
        cls,
        link: str,
        download_dir: str,
        task_id: int,
        number_of_dates: int,
        verbose: str = True,
    ) -> None:
        """
        Downloads file from link using PySmartDL
        :param link: link that needs to be downloaded
        :param download_dir: A relative path to the download directory
        :param task_id: ID of thread
        :verbose: Determines if download status is printed to console
        """
        global global_progress_tracker
        if (len(global_progress_tracker) == 1) and (number_of_dates > 1):
            global_progress_tracker = global_progress_tracker * number_of_dates
        downloader_obj = SmartDL(link, download_dir, progress_bar=False)
        downloader_obj.start(blocking=False)

        global global_final_download_size
        global_final_download_size += downloader_obj.get_final_filesize()

        while not downloader_obj.isFinished():
            if verbose:
                lock = Lock()
                lock.acquire()

                global_progress_tracker[
                    task_id
                ] = cls._get_individual_download_stats(downloader_obj)

                cls._update_global_download_size()
                cls._update_global_download_speed()
                cls._update_global_progress_percentage()

                # Prevents the time output from being constantly updated
                # to the point of being unreadable
                if random.randint(1, 10) <= 10:
                    cls._update_global_eta()

                progress_bar = (
                    f"> {humanize.naturalsize(global_download_size, binary=True, format=_TWO_DECIMALS)} / "
                    f"{humanize.naturalsize(global_final_download_size, binary=True, format=_TWO_DECIMALS)} "
                    f"@ {humanize.naturalsize(global_download_speed, binary=True)}/s {cls._create_progress_bar(size=20)} "
                    f"[ {global_progress_percentage:.2f}% ] "
                    f"{humanize.precisedelta(global_eta)} left                                \r"
                ).replace("i", "", 3)

                sys.stdout.write(progress_bar)

                sys.stdout.flush()
                lock.release()
            time.sleep(0.1)
